import json
import inspect
import sys
import dataclasses
from typing import Any, NewType, Dict, Union, Callable, List, Tuple, Optional, Type
try:
    from typing import Literal  # type: ignore
except(ImportError):
    # Literal polyfill
    import types

    class _LiteralMeta(type):
        @classmethod
        def __getitem__(cls, key):
            tp = type(key[0] if isinstance(key, tuple) else key)

            def _make_cls(ns):
                class Origin:
                    pass

                origin = Origin()
                setattr(origin, '_name', 'Literal')
                ns['__origin__'] = origin
                ns['__args__'] = key
                return ns

            return types.new_class('Literal', (tp,), {}, _make_cls)

    class Literal(metaclass=_LiteralMeta):
        pass
AllArguments = NewType('AllArguments', Dict[str, Any])
_empty = object()


class DefaultFactory:
    def __init__(self, factory):
        self.factory = factory

    def __repr__(self):
        return repr(self.factory())

    def __str__(self):
        return str(self.factory())

    def __call__(self):
        return self.factory()

    def __eq__(self, other):
        if not isinstance(other, DefaultFactory):
            return False
        return self._get_comp_value(self) == self._get_comp_value(other)

    @staticmethod
    def _get_comp_value(value):
        value = value.get_default()
        if isinstance(value, (int, str, float, bool)):
            return value
        return json.dumps(value(), sort_keys=True)

    def __hash__(self):
        return hash(self._get_comp_value(self))

    def get_default(self):
        value = self()
        if isinstance(value, (int, str, float, bool)):
            return value
        return self

    @staticmethod
    def get_factory(value):
        if isinstance(value, DefaultFactory):
            return value
        if value == _empty or value == inspect._empty:
            return None
        return DefaultFactory(lambda: value)


@dataclasses.dataclass
class Parameter:
    name: Optional[str]
    type: Optional[Type]
    help: str = ''
    children: List['Parameter'] = dataclasses.field(default_factory=list)
    default_factory: Optional[Callable[[], Any]] = None
    choices: Optional[List[Any]] = None
    argument_type: Optional[Any] = None
    _argument_name: Optional[Tuple[str]] = None

    def walk(self, fn: Callable[['ParameterWithPath', List[Any]], Any]):
        def _walk(e, parent):
            e = ParameterWithPath(e, parent)
            new_children = []
            for p in e.children:
                result = _walk(p, e)
                if result is not None:
                    if isinstance(result, ParameterWithPath):
                        result = result.parameter
                    new_children.append(result)
            return fn(e, children=new_children)
        result = _walk(self, None)
        if isinstance(result, ParameterWithPath):
            return result.parameter
        return result

    def enumerate_parameters(self):
        def _enumerate(e, parent):
            e = ParameterWithPath(e, parent)
            yield e
            for x in e.children:
                for y in _enumerate(x, e):
                    yield y
        return _enumerate(self, None)

    def __getitem__(self, name):
        for x in self.children:
            if x.name == name:
                return x
        raise IndexError(f'Element {name} not found')

    def __contains__(self, name):
        for x in self.children:
            if x.name == name:
                return True
        return False

    def find(self, name: str) -> Optional['Parameter']:
        for x in self.children:
            if x.name == name:
                return x
        return None

    @property
    def default(self):
        if self.default_factory is None:
            return None
        if isinstance(self.default_factory, DefaultFactory):
            return self.default_factory.get_default()
        return self.default_factory()

    def __str__(self):
        result = f'{self.name} [{self.type}]\n'
        for x in self.children:
            result += '  ' + str(x).replace('\n', '\n  ')
        return result.strip('\n ')

    def replace(self, **kwargs):
        return dataclasses.replace(self, **kwargs)


@dataclasses.dataclass
class ParameterWithPath:
    parameter: Parameter
    parent: Optional['ParameterWithPath'] = None

    @property
    def name(self):
        return self.parameter.name

    @property
    def type(self):
        return self.parameter.type

    @property
    def argument_type(self):
        return self.parameter.argument_type

    @property
    def children(self):
        return self.parameter.children

    @property
    def default_factory(self):
        return self.parameter.default_factory

    @property
    def default(self):
        return self.parameter.default

    @property
    def choices(self):
        return self.parameter.choices

    @property
    def help(self):
        return self.parameter.help

    @property
    def full_name(self):
        if self.parent is not None and self.parent.name is not None:
            return self.parent.full_name + '.' + self.name
        return self.name

    def find(self, name):
        child = self.parameter.find(name)
        if child is not None:
            return ParameterWithPath(child, self)
        return None

    @property
    def argument_name(self):
        if self.parameter._argument_name is not None:
            return self.parameter._argument_name[0]
        if self.parent is not None and self.parent.argument_name is not None:
            return self.parent.argument_name + '_' + self.name
        return self.name

    def replace(self, **kwargs):
        return ParameterWithPath(self.parameter.replace(**kwargs), self.parent)


class Runtime:
    def add_parameter(self, argument_name: str,
                      argument_type: Type, required: bool = True,
                      help: str = None,
                      default: Any = _empty, choices: Optional[List[Any]] = None):
        raise NotImplementedError('add_parameter is not implemented')

    def read_defaults(self, parameters: Parameter) -> Parameter:
        raise NotImplementedError('read_defaults is not implemented')


class Handler:
    def preprocess_parameter(self, parameter: ParameterWithPath) -> Tuple[bool, ParameterWithPath]:
        return False, parameter

    def parse_value(self, parameter: ParameterWithPath, value: Any) -> Tuple[bool, Any]:
        return False, value

    def bind(self, parameter: ParameterWithPath, args: Dict[str, Any], children: List[Tuple[Parameter, Any]]) -> Tuple[bool, Any]:
        return False, args

    def add_parameter(self, parameter: ParameterWithPath, parser: Runtime) -> bool:
        return False


class _ConditionalTypeMeta(type):
    def __new__(cls, name, bases, ns, prefix=True):
        """Create new typed dict class object.
        This method is called when ConditionalType is subclassed,
        or when ConditionalType is instantiated. This way
        ConditionalType supports all three syntax forms described in its docstring.
        Subclasses and instances of ConditionalType return actual dictionaries.
        """
        for base in bases:
            if type(base) is not _ConditionalTypeMeta:
                raise TypeError('cannot inherit from both a ConditionalType type '
                                'and a non-ConditionalType base class')

        annotations = {}
        own_annotations = ns.get('__annotations__', {})
        for base in bases:
            annotations.update(base.__dict__.get('__annotations__', {}))

        annotations.update(own_annotations)
        tp = Union.__getitem__(tuple(annotations.values()))
        setattr(tp, '__conditional_map__', annotations)
        setattr(tp, '__conditional_prefix__', prefix)
        return tp

    def __subclasscheck__(cls, other):
        # Typed dicts are only for static structural subtyping.
        raise TypeError('ConditionalType does not support instance and class checks')

    __instancecheck__ = __subclasscheck__


def ConditionalType(typename, fields=None, *, prefix: bool = True, **kwargs):
    """ConditionalType allows aparse to condition its choices for a
    specific parameter based on an argument.
    Usage::
        class Model(ConditionalType):
            gpt2: GPT2
            resnet: ResNet
    The type info can be accessed via the Model.__annotations__ dict,
    and the Model.__required_keys__ and Model.__optional_keys__ frozensets.
    ConditionalType supports two additional equivalent forms:
        Model = ConditionalType('Model', gpt2=GPT2, resnet=ResNet)
        Model = ConditionalType('Model', dict(gpt2=GPT2, resnet=ResNet))
    The class syntax is only supported in Python 3.6+, while two other
    syntax forms work for Python 2.7 and 3.2+
    """
    if fields is None:
        fields = kwargs
    elif kwargs:
        raise TypeError("ConditionalType takes either a dict or keyword arguments,"
                        " but not both")

    ns = {'__annotations__': dict(fields)}
    try:
        # Setting correct module is necessary to make typed dict classes pickleable.
        ns['__module__'] = sys._getframe(1).f_globals.get('__name__', '__main__')
    except (AttributeError, ValueError):
        pass

    return _ConditionalTypeMeta(typename, (), ns, prefix=prefix)


_ConditionalType = type.__new__(_ConditionalTypeMeta, 'ConditionalType', (), {})
setattr(ConditionalType, '__mro_entries__', lambda bases: (_ConditionalType,))



import inspect
from typing import Any
from functools import reduce
import dataclasses
import copy
from .core import Parameter, _empty, DefaultFactory


def get_path(obj, path, default=_empty, _current_path=None):
    if obj == _empty:
        raise IndexError(f'Could not find path {_current_path}.')
    if path == '':
        return obj
    curr, *rest = path.split('.')
    _current_path = curr if _current_path is None else f'{_current_path}.{curr}'
    if hasattr(obj, '__getitem__'):
        if curr in obj:
            obj = obj[curr]
        elif default != _empty:
            return default
        else:
            raise IndexError(f'Could not find path {_current_path}.')
    else:
        if hasattr(obj, curr):
            obj = getattr(obj, curr)
        elif default != _empty:
            return default
        else:
            raise IndexError(f'Could not find path {_current_path}.')

    return get_path(obj, '.'.join(rest), default=default, _current_path=_current_path)


def prefix_parameter(parameter, prefix, container_type=None):
    parameter = copy.copy(parameter)
    has_container = True
    if parameter.name is not None:
        root = Parameter(name=None, type=dict, children=[parameter])
        parameter = root
        has_container = False

    def prefix_single(x, name):
        parent = Parameter(name=name, type=container_type)
        parent.children = x.children
        x.children = [parent]
        return x

    parts = reversed([x for x in prefix.split('.') if x != ''])
    parameter = reduce(prefix_single, parts, parameter)
    if not has_container:
        parameter = parameter.children[0]
    return parameter


def merge_parameter_trees(*args):
    # Fix single parameters
    def _fix_single_parameter(x):
        if x.name is not None:
            root = Parameter(name=None, type=dict)
            root.children = [x]
            return root
        return x

    parameters = list(map(_fix_single_parameter, (x for x in args if x is not None)))
    return parameters[0].replace(children=[y for x in parameters for y in x.children])


def get_parameters(obj: Any) -> Parameter:
    generated = set()
    root = Parameter(name=None, type=dict)

    def collect_parameters(cls, parent, parent_name=None):
        parameters = inspect.signature(cls.__init__).parameters
        calls_parent = False
        for p in parameters.values():
            if p.name == 'self':
                continue
            if p.kind == inspect.Parameter.VAR_KEYWORD:
                for base in cls.__bases__:
                    calls_parent = True
            if p.kind == inspect.Parameter.KEYWORD_ONLY or p.kind == inspect.Parameter.POSITIONAL_OR_KEYWORD:
                full_name = f'{parent_name}.{p.name}' if parent_name is not None else p.name
                assert full_name not in generated, f'{full_name} was already generated'
                if dataclasses.is_dataclass(cls) and isinstance(p.default, dataclasses._HAS_DEFAULT_FACTORY_CLASS):
                    default_factory = DefaultFactory(cls.__dataclass_fields__[p.name].default_factory)
                else:
                    default_factory = DefaultFactory.get_factory(p.default)

                param = Parameter(p.name, p.annotation, default_factory=default_factory)
                generated.add(full_name)
                if dataclasses.is_dataclass(p.annotation):
                    param.children.extend(collect_parameters(p.annotation, param, full_name))
                yield param
        if calls_parent:
            for p in collect_parameters(base, parent, parent_name):
                yield p

    if inspect.isclass(obj):
        params = list(collect_parameters(obj, root, None))
    else:
        params = []
        parameters = inspect.signature(obj).parameters
        for p in parameters.values():
            if p.kind == inspect.Parameter.KEYWORD_ONLY or p.kind == inspect.Parameter.POSITIONAL_OR_KEYWORD:
                default_factory = DefaultFactory.get_factory(p.default)
                param = Parameter(
                    p.name,
                    p.annotation,
                    default_factory=default_factory,
                )
                if dataclasses.is_dataclass(p.annotation):
                    # Hierarchical arguments
                    param.children.extend(collect_parameters(p.annotation, param, param.name))
                params.append(param)
    root = root.replace(children=params)
    return root


# def walk_multiple(callback: Callable[[List[ParameterWithPath], List[ParameterWithPath]], Optional[ParameterWithPath]],
#                   parameter: Parameter, *others: Parameter) -> Parameter:
#     def _walk(params):
#         children = []
#         for p in params[0].children:
#             matching_children = [x.find(p.name) for x in params[1:]]
#             matching_children = [ParameterWithPath(x, par) for x, par in zip(matching_children, params) if x is not None]
#             child = _walk([ParameterWithPath(p, params[0])] + matching_children)
#             if child is not None:
#                 if isinstance(child, ParameterWithPath):
#                     child = child.parameter
#                 children.append(child)
# 
#         return callback(params, children)
#     return _walk([ParameterWithPath(x, p) for x, p in zip([parameter] + list(others), [None] * (1 + len(others)))]).parameter


# def is_type_compatible(annotation, type2):
#     if annotation == type2:
#         return True
#     meta_name = getattr(getattr(annotation, '__origin__', None), '_name', None)
#     if meta_name == 'Literal':
#         arg_type = type(annotation.__args__[0])
#         return arg_type == type2
#     if meta_name == 'Union':
#         return any(is_type_compatible(x, type2) for x in annotation.__args__)
#     return False


def consolidate_parameter_tree(parameters: Parameter, soft_defaults: bool = False) -> Parameter:
    def _merge(param, other):
        default_factory = param.default_factory
        choices = param.choices
        if param.type != other.type:
            raise Exception(f'Could not merge parameter {param.full_name} with different types: {param.type}, {other.type}')

        if param.default_factory is not None and other.default_factory is not None and other.default_factory != param.default_factory:
            if not soft_defaults:
                raise Exception(f'There are conflicting values for {param.argument_name}, [{other.default}, {param.default}]')
            else:
                default_factory = other.default_factory

        if param.default_factory is None and other.default_factory is not None:
            # Copy default value
            default_factory = other.default_factory

        if other.choices is not None or choices is not None:
            # Update choices in the literal
            if other.choices is None or choices is None:
                choices = list(other.choices or []) + list(choices or [])
            else:
                choices = sorted(set(other.choices).intersection(set(choices)))
        return param.replace(default_factory=default_factory, choices=choices)

    def _walk(param, children):
        new_children = []
        children_map = dict()
        for c in children:
            if c.name not in children_map:
                children_map[c.name] = len(new_children)
                new_children.append(c)
                continue

            first_c_i = children_map[c.name]
            new_children[first_c_i] = _merge(new_children[first_c_i], c)

        return param.replace(children=new_children)
    return parameters.walk(_walk)


def ignore_parameters(parameters, ignore):
    def _call(x, children):
        if x.full_name in ignore:
            return None
        return x.replace(children=children)
    return parameters.walk(_call)



from typing import Type
from .core import Handler, AllArguments, ParameterWithPath, Runtime, _empty
from ._lib import register_handler, preprocess_parameter
from .utils import get_parameters
from .utils import prefix_parameter, merge_parameter_trees


@register_handler
class DefaultHandler(Handler):
    def preprocess_parameter(self, param: ParameterWithPath):
        if len(param.children) > 0:
            return True, param.parameter
        if param.type is None:
            return True, None

        choices = None
        meta_name = getattr(getattr(param.type, '__origin__', None), '_name', None)
        default = param.default_factory() if param.default_factory is not None else None
        arg_type = None
        if meta_name == 'Literal':
            arg_type = type(param.type.__args__[0])
            choices = param.type.__args__
        elif meta_name == 'Union':
            tp = set((x for x in param.type.__args__ if not isinstance(None, x)))
            if str in tp:
                arg_type = str
            if int in tp:
                arg_type = int
            if float in tp:
                arg_type = float
        elif hasattr(param.type, 'from_str') and callable(param.type.from_str):
            arg_type = str
        elif param.type in [int, float, str, bool]:
            arg_type = param.type
        elif isinstance(default, bool):
            arg_type = bool
        if arg_type is not None:
            return True, param.replace(argument_type=arg_type, choices=choices)
        return False, None

    def add_parameter(self, param, runtime: Runtime):
        if len(param.children) > 0 or param.argument_type not in {str, float, bool, int}:
            return True

        argument_name = param.argument_name
        if param.argument_type == bool:
            negative_option = argument_name
            if negative_option.startswith('use-'):
                negative_option = negative_option[len('use-'):]
            argument_name += '/' + f'--no-{negative_option}'
        required = param.default_factory is None
        runtime.add_parameter(param.argument_name, param.argument_type,
                              required=required, help=param.help,
                              default=param.default if param.default_factory is not None else _empty,
                              choices=param.choices)
        return True


@register_handler
class AllArgumentsHandler(Handler):
    def add_parameter(self, param, runtime, *args, **kwargs):
        if param.type == AllArguments:
            return True
        return False

    def bind(self, param, args, children):
        if param.type == AllArguments:
            value = {k: v for k, v in args.items() if k != '_aparse_parameters'}
            return True, value
        return False, args

    def preprocess_parameter(self, param):
        if param.type == AllArguments:
            return True, param.replace(argument_type=AllArguments, children=[])
        return False, param


@register_handler
class SimpleListHandler(Handler):
    def _list_type(self, tp: Type):
        if getattr(tp, '__origin__', None) == list:
            tp = tp.__args__[0]
            if tp in (int, str, float):
                return tp
        return None

    def preprocess_parameter(self, parameter):
        if self._list_type(parameter.type) is not None:
            return True, parameter.replace(argument_type=str)
        return False, parameter

    def parse_value(self, parameter, value):
        list_type = self._list_type(parameter.type)
        if list_type is not None and isinstance(value, str):
            return True, list(map(list_type, value.split(',')))
        return False, value


@register_handler
class FromStrHandler(Handler):
    def _does_handle(self, tp: Type):
        return hasattr(tp, 'from_str')

    def preprocess_parameter(self, parameter):
        if self._does_handle(parameter.type):
            return True, parameter.replace(argument_type=str)
        return False, parameter

    def parse_value(self, parameter, value):
        if self._does_handle(parameter.type) and isinstance(value, str):
            return True, parameter.type.from_str(value)
        return False, value


@register_handler
class ConditionalTypeHandler(Handler):
    @staticmethod
    def _does_handle(tp: Type):
        meta_name = getattr(getattr(tp, '__origin__', None), '_name', None)
        if meta_name == 'Union':
            return hasattr(tp, '__conditional_map__')
        return False

    def preprocess_parameter(self, parameter):
        if self._does_handle(parameter.type):
            return True, parameter.replace(
                argument_type=str,
                choices=list(parameter.type.__conditional_map__.keys()))
        return False, parameter

    def before_parse(self, root, parser, kwargs):
        result = []
        for param in root.enumerate_parameters():
            if self._does_handle(param.type):
                key = kwargs.get(param.argument_name, param.default)
                tp = param.type.__conditional_map__.get(key, None)
                if tp is not None:
                    parameter = get_parameters(tp).walk(preprocess_parameter)
                    parameter = parameter.replace(name=param.name, type=tp)
                    if not param.type.__conditional_prefix__:
                        parameter = parameter.replace(_argument_name=(None,))
                    if param.parent is not None and param.parent.full_name is not None:
                        parameter = prefix_parameter(parameter, param.parent.full_name)
                    result.append(parameter)
        if len(result) > 0:
            result = merge_parameter_trees(*result)
            return result
        return None

    def after_parse(self, root, argparse_args, kwargs):
        for param in root.enumerate_parameters():
            if ConditionalTypeHandler._does_handle(param.type):
                pass
        return kwargs



import click
from aparse.core import Parameter, Runtime, DefaultFactory
from aparse._lib import preprocess_parameter as _preprocess_parameter
from aparse._lib import add_parameters as _add_parameters
from aparse._lib import handle_before_parse as _handle_before_parse
from aparse._lib import parse_arguments_manually as _parse_arguments_manually
from aparse._lib import bind_parameters as _bind_parameters
from aparse._lib import handle_after_parse as _handle_after_parse
from aparse.utils import _empty, get_parameters as _get_parameters
from aparse.utils import merge_parameter_trees
# from click import *  # noqa: F403, F401


class ClickRuntime(Runtime):
    def __init__(self, fn, soft_defaults=False):
        self.fn = fn
        self.soft_defaults = soft_defaults
        self._parameters = None
        self._after_parse_callbacks = []

    def add_parameter(self, argument_name, argument_type, required=True,
                      help='', default=_empty, choices=None):
        if choices is not None:
            argument_type = click.Choice(choices, case_sensitive=True)
        if argument_type is not None:
            params = self._get_params()
            existing_action = {x.name: x for x in params}.get(argument_name, None)
            if existing_action is not None:
                params.remove(existing_action)

            self.fn = click.option(f'--{argument_name.replace("_", "-")}', type=argument_type,
                                   required=required, default=default if default != _empty else None,
                                   show_default=default != _empty, help=help,
                                   show_choices=choices is not None)(self.fn)

    def _get_params(self):
        return getattr(self.fn, 'params', getattr(self.fn, '__click_params__', []))

    @staticmethod
    def _get_click_type(click_type):
        if click_type == click.INT:
            return int
        elif click_type == click.STRING:
            return str
        elif click_type == click.FLOAT:
            return float
        elif click_type == click.BOOL:
            return bool
        elif isinstance(click_type, click.Choice):
            return type(click_type.choices[0] if len(click_type.choices) > 0 else None)
        return click_type

    def read_defaults(self, parameters: Parameter):
        params = self._get_params()
        if len(params) == 0:
            return parameters
        param_map = {x.name: x for x in params}

        def map(param, children):
            choices = param.choices
            default_factory = param.default_factory
            if param.name is None or param.type is None or len(param.children) > 0:
                return param.replace(children=children)

            existing_action = param_map.get(param.argument_name, None)
            if existing_action is None:
                return param.replace(children=children)

            default_factory = DefaultFactory.get_factory(existing_action.default)
            if existing_action.required:
                default_factory = None
            if isinstance(existing_action.type, click.Choice):
                choices = existing_action.type.choices
            argument_type = self._get_click_type(existing_action.type)
            return param.replace(choices=choices, default_factory=default_factory,
                                 argument_type=argument_type, children=children)
        return parameters.walk(map)

    def add_parameters(self, parameters: Parameter):
        _add_parameters(parameters, self, soft_defaults=self.soft_defaults)

        if self._parameters is not None:
            parameters = merge_parameter_trees(self._parameters, parameters)
        self._parameters = parameters


def _get_command_class(cls=None):
    class AparseClickCommand(cls or click.core.Command):
        runtime = None

        def __init__(self, *args, **kwargs):
            super().__init__(*args, **kwargs)

        def invoke(self, ctx):
            kwargs = ctx.params
            kwargs = _bind_parameters(self.runtime._parameters, kwargs)
            kwargs = _handle_after_parse(self.runtime._parameters, ctx.params, kwargs, self.runtime._after_parse_callbacks)
            ctx.params = kwargs
            return super().invoke(ctx)
    return AparseClickCommand


def command(name=None, cls=None, before_parse=None, after_parse=None, soft_defaults=False, **kwargs):
    cls = _get_command_class(cls)
    _wrap = click.command(name=name, cls=cls, **kwargs)

    def wrap(fn):
        root_param = _get_parameters(fn).walk(_preprocess_parameter)
        runtime = ClickRuntime(fn, soft_defaults=soft_defaults)
        cls.runtime = runtime
        if after_parse is not None:
            runtime._after_parse_callbacks.append(after_parse)
        runtime.add_parameters(root_param)
        kwargs = _parse_arguments_manually()
        new_param = _handle_before_parse(runtime, root_param, kwargs, [before_parse])
        if new_param is not None:
            runtime.add_parameters(new_param)

        fn = runtime.fn
        fn = _wrap(fn)
        return fn

    return wrap


def _get_group_class(cls=None):
    class Group(cls or click.core.Group):
        def command(self, name=None, cls=None, **kwargs):
            if callable(name):
                return self.command()(name)

            _wrap = command(name=name, cls=cls, **kwargs)

            def wrap(fn):
                fn = _wrap(fn)
                self.add_command(fn)
                return fn
            return wrap
    return Group


def group(name=None, cls=None, **attrs):
    return click.group(name=name, cls=_get_group_class(cls), **attrs)



import sys
from typing import List, Dict, Any, Tuple
import dataclasses
from .core import Parameter, ParameterWithPath, Handler, Runtime, DefaultFactory
from .utils import merge_parameter_trees, consolidate_parameter_tree
from .utils import ignore_parameters


handlers: List[Handler] = []


def register_handler(handler):
    handlers.insert(0, handler())
    return handler


def preprocess_parameter(param: ParameterWithPath, children):
    handled = False
    param = param.replace(children=children)
    if param.name is not None:
        for h in handlers:
            handled, param = h.preprocess_parameter(param)
            if handled:
                break
    return param


def parse_arguments_manually(args=None, defaults=None):
    kwargs = dict(**defaults) if defaults is not None else dict()
    if args is None:
        # args default to the system args
        args = sys.argv[1:]
    else:
        # make sure that args are mutable
        args = list(args)

    for nm, val in zip(args, args[1:]):
        if nm.startswith('--') and '=' not in nm:
            if val.startswith('--'):
                val = True
            kwargs[nm[2:].replace('-', '_')] = val
        elif val.startswith('--'):
            kwargs[val[2:]] = True
    else:
        for nm in args:
            if nm.startswith('--') and '=' in nm:
                nm, val = nm[2:].split('=', 1)
                kwargs[nm[2:].replace('-', '_')] = val
    return kwargs


def handle_before_parse(runtime: Runtime, parameters: Parameter, kwargs: Dict[str, str], callbacks=None):
    added_params: List[Parameter] = []
    for bp in [getattr(h, 'before_parse', None) for h in reversed(handlers)] + (callbacks or []):
        if bp is None:
            continue
        np = bp(merge_parameter_trees(*([parameters] + added_params)), runtime, kwargs)
        if np is not None:
            added_params.append(np)

    if len(added_params) > 0:
        return merge_parameter_trees(*added_params).walk(preprocess_parameter)
    return None


def handle_after_parse(parameters: Parameter, arguments: Dict[str, Any], kwargs: Dict[str, Any], callbacks=None):
    for ap in [getattr(h, 'after_parse', None) for h in reversed(handlers)] + (callbacks or []):
        if ap is None:
            continue
        kwargs = ap(parameters, arguments, kwargs)
    return kwargs


def set_defaults(parameters: Parameter, defaults: Dict[str, Any]):
    defaults = dict(**defaults)

    def _call(param, children):
        default_factory = param.default_factory
        if param.full_name in defaults:
            default = defaults.pop(param.full_name)
            default_factory = DefaultFactory.get_factory(default)
        return param.replace(
            children=children,
            default_factory=default_factory)
    return parameters.walk(_call), defaults


def add_parameters(parameters: Parameter, runtime: Runtime,
                   defaults: Dict[str, Any] = None,
                   soft_defaults: bool = False):
    if defaults is not None:
        parameters, left_defaults = set_defaults(parameters, defaults)
        if left_defaults:
            raise ValueError(f'Some default were not found: {list(left_defaults.keys())}')

    default_parameters = runtime.read_defaults(parameters)
    if defaults is not None:
        default_parameters = ignore_parameters(default_parameters, set(defaults.keys()))
    parameters = merge_parameter_trees(default_parameters, parameters)
    parameters = consolidate_parameter_tree(parameters, soft_defaults=soft_defaults)
    # TODO: implement default's type validation!
    # parameters = merge_parameter_defaults(parameters, default_parameters, soft_defaults=soft_defaults)

    for param in parameters.enumerate_parameters():
        for h in handlers:
            if h.add_parameter(param, runtime):
                break
        else:
            raise RuntimeError('There was no handler registered for adding arguments')


def bind_parameters(parameters: Parameter, arguments: Dict[str, Any]):
    def bind(parameter: ParameterWithPath, children: List[Tuple[Parameter, Any]]):
        was_handled = False
        value = arguments
        for h in handlers:
            was_handled, value = h.bind(parameter, arguments, children)
            if was_handled:
                break
        if was_handled:
            return parameter, value

        if len(children) > 0 or parameter.name is None:
            if dataclasses.is_dataclass(parameter.type):
                value = parameter.type(**{p.name: x for p, x in children})
            elif parameter.type == dict:
                value = {p.name: x for p, x in children if p.name is not None}
            else:
                raise RuntimeError(f'Aggregate type {parameter.type} is not supported')
        else:
            value = arguments[parameter.argument_name]
            if value == parameter.default:
                value = parameter.default_factory()
            else:
                was_handled = False
                for handler in handlers:
                    was_handled, value = handler.parse_value(parameter, value)
                    if was_handled:
                        break
        return parameter, value
    _, kwargs = parameters.walk(bind)
    return kwargs



from . import _lib
from .core import Handler, Parameter, ParameterWithPath, Literal, AllArguments  # noqa: F401
from .core import ConditionalType, DefaultFactory  # noqa: F401
from ._lib import register_handler  # noqa: F401
from .argparse import add_argparse_arguments  # noqa: F401
from . import _handlers  # noqa: F401

__all__ = ['Handler', 'Parameter', 'ParameterWithPath', 'Literal',
           'AllArguments', 'ConditionalType', 'register_handler',
           'add_argparse_arguments']

__version__ = "0.0.5"
del _lib
del _handlers



from typing import Dict, Set, Any, Optional, Callable
from functools import partial
from argparse import ArgumentParser, Namespace
from .core import Parameter, DefaultFactory, Runtime
from ._lib import add_parameters as _add_parameters
from ._lib import preprocess_parameter as _preprocess_parameter
from ._lib import handle_before_parse as _handle_before_parse
from ._lib import parse_arguments_manually as _parse_arguments_manually
from ._lib import bind_parameters as _bind_parameters
from .utils import _empty, merge_parameter_trees, prefix_parameter
from .utils import ignore_parameters, get_parameters as _get_parameters
from .utils import get_path as _get_path


class ArgparseRuntime(Runtime):
    def __init__(self, parser, soft_defaults: bool = False, defaults=None):
        self.parser = parser
        self._before_parse_callbacks = []
        if hasattr(parser, '_aparse_runtime'):
            old_soft_defaults, old_defaults, self._before_parse_callbacks = parser._aparse_runtime
            assert old_soft_defaults == soft_defaults
            if defaults is not None:
                old_defaults.update(defaults)
            defaults = old_defaults
        else:
            self.parser = self._hack_argparse(self.parser)
        self.soft_defaults = soft_defaults
        self.defaults = defaults or dict()
        self._save()

    def _save(self):
        setattr(self.parser, '_aparse_runtime', (self.soft_defaults, self.defaults, self._before_parse_callbacks))

    def register_before_parse_callback(self, callback):
        self._before_parse_callbacks.append(callback)

    def _hack_argparse(self, parser):
        super_parse_known_args = parser.parse_known_args

        def hacked_parse_known_args(args=None, namespace=None):
            defaults = dict(**self.defaults) if self.defaults is not None else dict()
            kwargs = _parse_arguments_manually(args, defaults)
            old_params = getattr(parser, '_aparse_parameters')
            callbacks = list(self._before_parse_callbacks)
            new_parameters = _handle_before_parse(self, old_params, kwargs, callbacks)
            if new_parameters is not None:
                self.add_parameters(new_parameters)
            result = super_parse_known_args(args, namespace)
            setattr(result[0], '_aparse_parameters', getattr(parser, '_aparse_parameters', None))
            return result

        setattr(parser, 'parse_known_args', hacked_parse_known_args)
        return parser

    def add_parameter(self, argument_name, argument_type, required=True,
                      help='', default=_empty, choices=None):

        existing_action = None
        for action in self.parser._actions:
            if action.dest == argument_name:
                existing_action = action
                break

        params = []
        if default != _empty:
            params.append('default: {default}')
        if required:
            params.append('required')
        help = f'{help} [{", ".join(params)}]' if len(params) > 0 else help

        # Find existing action
        if existing_action is not None:
            # We will update default
            existing_action.required = required
            existing_action.help = help
            existing_action.choices = choices
            self.parser.set_defaults(**{argument_name: default if default != _empty else None})
        else:
            arg_type = argument_type
            arg_name = argument_name.replace('_', '-')
            if arg_type == bool:
                assert '/' in arg_name
                true_arg_name, false_arg_name = arg_name.split('/', 2)
                if default != _empty:
                    self.parser.set_defaults(**{argument_name: default})
                self.parser.add_argument(f'--{true_arg_name}', dest=argument_name, action='store_true', help=f'{help} [default]' if default == True else help)  # noqa: E712
                self.parser.add_argument(f'--{false_arg_name}', dest=argument_name, action='store_false', help=f'{help} [default]' if default == False else help)  # noqa: E712
            else:
                self.parser.add_argument(f'--{arg_name}', type=arg_type, choices=choices, default=default if default != _empty else None,
                                         required=required,
                                         help=help)

    def read_defaults(self, parameters: Parameter) -> Parameter:
        def map(param, children):
            choices = param.choices
            default_factory = param.default_factory
            if param.name is None or param.type is None or len(param.children) > 0:
                return param.replace(children=children)

            for existing_action in self.parser._actions:
                if existing_action.dest == param.argument_name:
                    break
            else:
                return param.replace(children=children)

            default = existing_action.default
            default_factory = DefaultFactory.get_factory(default)
            if existing_action.required:
                default_factory = None
            argument_type = existing_action.type
            choices = existing_action.choices
            return param.replace(choices=choices, default_factory=default_factory,
                                 argument_type=argument_type, children=children)
        return parameters.walk(map)

    def add_parameters(self, parameters: Parameter):
        # Store parameters
        setattr(self.parser, '_aparse_parameters', merge_parameter_trees(getattr(self.parser, '_aparse_parameters', None), parameters))

        # Add parameters
        _add_parameters(parameters, runtime=self, soft_defaults=self.soft_defaults, defaults=self.defaults)


def _add_argparse_arguments(parameters: Parameter, parser: ArgumentParser,
                            defaults: Dict[str, Any] = None, prefix: str = None,
                            ignore: Optional[Set[str]] = None, soft_defaults: bool = False, _before_parse=None):
    runtime = ArgparseRuntime(parser, soft_defaults=soft_defaults, defaults=defaults)
    if prefix is not None:
        parameters = prefix_parameter(parameters, prefix, dict)
    if ignore is not None:
        parameters = ignore_parameters(parameters, ignore)
    if _before_parse is not None:
        runtime.register_before_parse_callback(_before_parse)
    runtime.add_parameters(parameters)
    return runtime.parser


def _bind_argparse_arguments(
        parameters: Parameter, argparse_args, ignore=None,
        after_parse: Optional[Callable[[Parameter, Dict[str, Any], Dict[str, Any]], Dict[str, Any]]] = None):
    args_dict = argparse_args.__dict__
    if '_aparse_parameters' in args_dict:
        args_dict = {k: v for k, v in args_dict.items()}
        parameters = args_dict.pop('_aparse_parameters')

    if ignore is not None:
        parameters = parameters.walk(lambda x, children:
                                     x.replace(children=children) if x.full_name not in ignore else None)

    kwargs = _bind_parameters(parameters, args_dict)
    if after_parse is not None:
        kwargs = after_parse(parameters, args_dict, kwargs)
    return kwargs


def _from_argparse_arguments(parameters: Parameter, function, argparse_args, *args, _ignore=None, _prefix: str = None, _after_parse=None, **kwargs):
    new_kwargs = _bind_argparse_arguments(parameters, argparse_args, ignore=set(kwargs.keys()).union(_ignore or []), after_parse=_after_parse)
    if _prefix is not None:
        new_kwargs = _get_path(new_kwargs, _prefix)
    new_kwargs.update(kwargs)
    return function(*args, **new_kwargs)


def add_argparse_arguments(
        _fn=None, *,
        ignore: Set[str] = None,
        before_parse: Callable[[ArgumentParser, Dict[str, Any]], ArgumentParser] = None,
        after_parse: Callable[[Namespace, Dict[str, Any]], Dict[str, Any]] = None):
    '''
    Extends function or class with "add_argparse_arguments", "from_argparse_arguments", and "bind_argparse_arguments" methods.
    "add_argparse_arguments" adds arguments to the argparse.ArgumentParser instance.
    "from_argparse_arguments" takes the argparse.Namespace instance obtained by calling parse.parse_args(), parses them and calls
        original function or constructs the class
    "bind_argparse_arguments" just parses the arguments into a kwargs dictionary, but does not call the original function. Instead,
        the parameters are returned.

    Arguments:
        ignore: Set of parameters to ignore when inspecting the function signature
        before_parse: Callback to be called before parser.parse_args()
        after_parse: Callback to be called before "from_argparse_arguments" calls the function and updates the kwargs.

    Returns: The original function extended with other functions.
    '''
    def wrap(fn):
        parameters = _get_parameters(fn).walk(_preprocess_parameter)
        if ignore is not None:
            parameters = ignore_parameters(parameters, ignore)
        setattr(fn, 'add_argparse_arguments', partial(_add_argparse_arguments, parameters, _before_parse=before_parse))
        setattr(fn, 'from_argparse_arguments', partial(_from_argparse_arguments, parameters, fn, _after_parse=after_parse))
        setattr(fn, 'bind_argparse_arguments', partial(_bind_argparse_arguments, parameters, after_parse=after_parse))
        return fn

    if _fn is not None:
        return wrap(_fn)
    return wrap
